import pygame
from pygame.locals import *
from lostsky.core.linalg import Vector2
from random import randint, choice
from sys import exit
from math import sin, cos
import os

def get_catalog():
    catalog = {'TS Test Trait': Purify(),
               'Airlift': AirLift(),
               'Dizzy Attack': DizzyAttack(),
               'Cat\'s Curse': CatsCurse(),
               'Illusion Field': IllusionField(),
               'Focused Movement': FocusedMovement(),
               'Teleport': Teleport(),
               'Double Action': DoubleAction(),
               'Summon Doll': SummonDoll(),
               'Magic Fortress': MagicFortress(),
               'Field Medic': FieldMedic(),
               'Dispel History': DispelHistory(),
               'Unleash Grimoire': UnleashGrimoire(),
               'Purify': Purify(),
               'Self Destruct':SelfDestruct(),
                }
    return catalog


class Skill(object):

    def __init__(self, name, desc, sc_minimum, sc_cost):

        """
        Base class for all Trait Skills

        # Inputs:   name - name of trait
                    desc - description of trait
                    sc_minimum - minimum SC needed to use this trait
                    sc_cost - SC cost used in this trait

        """
        self.name = name
        self.desc = desc
        self.sc_minimum = sc_minimum
        self.sc_cost = sc_cost

    def player_interface(self, unit):

        """
        Base class: no player interface available

        """
        pass


    def check_usability(self, unit):
        """
        Checks whether a unit can use a certain skill. Default is to check for enough spirit charge.
        """
        if unit.spirit >= self.sc_cost and unit.spirit >= self.sc_minimum:
            return True

    def check_criteria(self, unit, target):

        """
        Checks whether a given target is a valid selection for this skill
        """
        return False

    def generate_skill_range(self, minimum_range, maximum_range):

        """
        # Function Name: get_attack_range
        # Purpose: generates the valid range of attack tiles for a spell
        """

        validattacks = []

        # generates the valid moveset for 1st quadrant
        # General Pattern:
        #
        #  X123
        #  123
        #  23
        #  3
        #

        for diagonal_row_num in xrange(minimum_range, maximum_range+1):
            for index in xrange(0, diagonal_row_num+1):
                validattacks.append((diagonal_row_num-index, index))

        # generates the valid moveset for 2nd, 3rd, 4th quadrants
        for coord in list(validattacks):

            validattacks.append((-coord[0], coord[1]))
            validattacks.append((-coord[0], -coord[1]))
            validattacks.append((coord[0], -coord[1]))

        # Removes duplicates
        return list(set(validattacks))


    def generate_move_range(self, unit, maximum_range):
        """
        Generates a movement range based on a given maximum number of tiles from the target. Used for teleport spells.
        """


        move_range = self.generate_skill_range(1, maximum_range)

        for destination in list(move_range):
            if unit.map.check_occupancy(destination):
                move_range.remove(destination)

        return move_range


    def plot_skill_range(self, unit, skill_range, tile_type):
        """
        # Function Name: plot_attack
        # Purpose: Plots the valid attack range for the skill
        # Inputs: input: selection_range -
        """

        for tile in skill_range:
            if tile_type == 'attack':
                unit.map.engine.surface.blit(unit.map.engine.attack_tile, (Vector2(tile)*35+unit.location_pixel-unit.map.screen_shift*35))
            elif tile_type == 'healing':
                unit.map.engine.surface.blit(unit.map.engine.heal_tile, (Vector2(tile)*35+unit.location_pixel-unit.map.screen_shift*35))
            elif tile_type == 'move':
                unit.map.engine.surface.blit(unit.map.engine.move_tile, (Vector2(tile)*35+unit.location_pixel-unit.map.screen_shift*35))



    def select_unit(self, unit, skill_range, tile_type):

        menu_flag = True
        unit.map.framenum = 0

        while menu_flag:

            arrowkeys = False
            if unit.map.framenum == 10:
                unit.map.framenum = 0
            unit.map.framenum += 1

            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                        unit.map.cursor_arrows(event)
                        arrowkeys = True
                        # Resets the frame counter
                        unit.map.framenum = 0
                    if event.key == K_x:
                        menu_flag = False
                        unit.map.center_on(unit)
                    if event.key == K_z or event.key == K_RETURN:

                        # If the Z key is pressed, it will calculate the vector between the current position
                        # of the cursor and the unit's location. If that vector is part of the valid moves,
                        # the unit is moved there.
                        target_name = unit.map.cursor_key_search()
                        if target_name:
                            target = unit.map.all_units_by_name[target_name]
                            if self.check_criteria(unit, target):
                                return target

            # if there is not a tap detected, check if the key is being held down
            if arrowkeys == False and unit.map.framenum == 9:
                unit.map.cursor_arrows_hold()

            # Plots the valid attack tiles
            if menu_flag:

                pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f, %1.0f) - Shift (%1.0f, %1.0f)"
                %(unit.map.cursor_pos.x, unit.map.cursor_pos.y, unit.map.screen_shift.x, unit.map.screen_shift.y))

                unit.map.render_background()
                self.plot_skill_range(unit, skill_range, tile_type)

                unit.map.render_all_units()
                unit.map.render_cursor()
                unit.map.engine.surface.blit(unit.map.engine.menu_board, (0, 490))
                unit.plot_stats()

                # Display for combat statistics
                # Checks if there is a unit underneath the cursor, and if there is, plots the unit's data.
                selected_stats = False
                selected_stats = unit.map.cursor_key_search()
                # Displays if there is a unit under your cursor, it is not on your team for attack class spells and or on your team for healing-class spells.
                if (selected_stats is not False and
                    ((unit.map.all_units_by_name[selected_stats].team != unit.team and
                      unit.spell_actions[unit.equipped].type == 'attack') or
                     (unit.map.all_units_by_name[selected_stats].team == unit.team and
                      unit.spell_actions[unit.equipped].type in ('healing', 'healingitem', 'support')))):

                    #Plots predictor system
                    unit.plot_predictor(unit.map.all_units_by_name[selected_stats])


                pygame.display.flip()
                unit.map.engine.clock.tick(60)

    def select_nearby_units(self, unit, team, max_range):

        """
        Selects nearby units:

        Input:  unit - unit using this skill
                same_team - 'ally' for identify nearby allies, 'enemy' for identify nearby enemies, 'both'
        """


        if unit.team == 1:
            if team == 'ally':
                search_team = unit.map.team1
            elif team == 'enemy':
                search_team = unit.map.team2
            elif team == 'both':
                search_team = unit.map.team1 + unit.map.team2
            else:
                search_team = []

        else:
            if team == 'ally':
                search_team = unit.map.team2
            elif team == 'enemy':
                search_team = unit.map.team1
            elif team == 'both':
                search_team = unit.map.team1 + unit.map.team2
            else:
                search_team = []

        nearby_units = []

        search_range_delta = self.generate_skill_range(0, max_range)
        search_range_absolute = [unit.location_tile + Vector2(delta) for delta in search_range_delta]


        for target in search_team:
            if target.location_tile in search_range_absolute:
                nearby_units.append(target)

        return nearby_units


    def select_move(self, unit, move_range):

        """
        # Function Name: move_loop
        # Purpose: The unit's second level move loop:
        #      Allows the player to select where they want to move a unit
        # Output: moved_flag = True if a selection has been made, False if one has not been made
        """

        menu_flag = True
        unit.map.framenum = 0
        while menu_flag:
            # Frame counter for holding down the keys to move
            if unit.map.framenum == 10:
                unit.map.framenum = 0
            unit.map.framenum += 1

            arrowkeys = False
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:

                    if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                        unit.map.cursor_arrows(event)
                        arrowkeys = True
                        # Resets the frame counter
                        unit.map.framenum = 0

                    if event.key == K_x:
                        menu_flag = False

                        unit.map.center_on(unit)
                        return False

                    if event.key == K_z or event.key == K_RETURN:

                        # If the Z key is pressed, it will calculate the vector between the current position
                        # of the cursor and the unit's location. If that vector is part of the valid moves,
                        # the unit is moved there.
                        delta_pos = unit.map.cursor_pos - unit.location_tile
                        new_pos = unit.map.cursor_pos
                        if tuple(delta_pos) in move_range and not unit.map.check_occupancy(new_pos):
                            return new_pos


            # if there is not a tap detected, check if the key is being held down
            if arrowkeys == False and unit.map.framenum == 9:
                unit.map.cursor_arrows_hold()

            if menu_flag:
                pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f, %1.0f) - Shift (%1.0f, %1.0f)"
                %(unit.map.cursor_pos.x, unit.map.cursor_pos.y, unit.map.screen_shift.x, unit.map.screen_shift.y))

                unit.map.render_background()

                self.plot_skill_range(unit, move_range, 'move')

                unit.map.render_all_units()
                unit.map.render_cursor()

                unit.map.engine.surface.blit(unit.map.engine.menu_board, (0, 490))

                text_from = unit.map.engine.bfont.render(unit.map.terrainmap[tuple(unit.location_tile)][0].name, True, (0, 0, 0))
                text_from_def = unit.map.engine.sfont.render("TDEF: %2.1f"%unit.map.terrainmap[tuple(unit.location_tile)][0].TDEF, True, (0, 0, 0))
                text_from_agl = unit.map.engine.sfont.render("AGL: "+str(unit.map.terrainmap[tuple(unit.location_tile)][0].aglmod), True, (0, 0, 0))
                text_from_acc = unit.map.engine.sfont.render("ACC: "+str(unit.map.terrainmap[tuple(unit.location_tile)][0].accmod), True, (0, 0, 0))


                text_to = unit.map.engine.bfont.render(unit.map.terrainmap[tuple(unit.map.cursor_pos)][0].name, True, (0, 0, 0))
                text_to_def = unit.map.engine.sfont.render("TDEF: %2.1f"%unit.map.terrainmap[tuple(unit.map.cursor_pos)][0].TDEF, True, (0, 0, 0))
                text_to_agl = unit.map.engine.sfont.render("AGL: "+str(unit.map.terrainmap[tuple(unit.map.cursor_pos)][0].aglmod), True, (0, 0, 0))
                text_to_acc = unit.map.engine.sfont.render("ACC: "+str(unit.map.terrainmap[tuple(unit.map.cursor_pos)][0].accmod), True, (0, 0, 0))

                # Plots the terrain data
                unit.map.engine.surface.blit(text_from, (450, 505))
                unit.map.engine.surface.blit(unit.map.engine.terrain_icon, (450, 523), (unit.map.terrainmap[tuple(unit.location_tile)][0].icon*50, 0, 50, 50))
                unit.map.engine.surface.blit(text_from_def, (450, 575))
                unit.map.engine.surface.blit(text_from_acc, (450, 587))
                unit.map.engine.surface.blit(text_from_agl, (450, 599))

                unit.map.engine.surface.blit(unit.map.engine.move_arrow, (565, 523))

                unit.map.engine.surface.blit(text_to, (810-text_to.get_width(), 505))
                unit.map.engine.surface.blit(unit.map.engine.terrain_icon, (760, 523), (unit.map.terrainmap[tuple(unit.map.cursor_pos)][0].icon*50, 0, 50, 50))
                unit.map.engine.surface.blit(text_to_def, (810-text_to_def.get_width(), 575))
                unit.map.engine.surface.blit(text_to_acc, (810-text_to_acc.get_width(), 587))
                unit.map.engine.surface.blit(text_to_agl, (810-text_to_agl.get_width(), 599))

                # Checks if there is a unit underneath the cursor, and if there is, plots the unit's data.
                selected = False
                selected = unit.map.cursor_key_search()
                if selected is not False:
                    unit.map.all_units_by_name[selected].plot_stats()

                pygame.display.flip()
                unit.map.engine.clock.tick(60)


    def confirm_proximity_action(self, unit, target_list, skill_range, tile_type):


        text_confirm_action = unit.map.engine.sfont.render("Confirm action, \"%s\"?"%self.name, True, (0, 0, 0))
        if target_list:
            text_yes = unit.map.engine.sfont.render('YES', True, (0, 0, 0) )
        else:
            text_yes = unit.map.engine.sfont.render('YES', True, (100, 100, 100) )


        text_no = unit.map.engine.sfont.render('NO', True, (0, 0, 0) )

        confirm_action = True

        menu_flag = True
        while menu_flag:
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:

                    if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                        confirm_action = not confirm_action
                    if event.key == K_x:
                        unit.map.center_on(unit)
                        return False

                    if event.key == K_z or event.key == K_RETURN:
                        # Can only confirm if targets are actually present
                        if target_list:
                            return confirm_action
                        elif not confirm_action:
                            return False
                        else:
                            pass

            if menu_flag:
                pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f, %1.0f) - Shift (%1.0f, %1.0f)"
                %(unit.map.cursor_pos.x, unit.map.cursor_pos.y, unit.map.screen_shift.x, unit.map.screen_shift.y))

                unit.map.render_background()


                self.plot_skill_range(unit, skill_range, tile_type)
                unit.map.render_all_units()


                for target in target_list:
                    unit.map.cursor_pos = target.location_tile
                    unit.map.render_cursor()

                unit.map.engine.surface.blit(unit.map.engine.menu_board, (0, 490))


                # Plots the terrain data

                unit.map.engine.surface.blit(text_confirm_action, (450, 505))
                unit.map.engine.surface.blit(text_yes, (460, 520))
                unit.map.engine.surface.blit(text_no, (460, 535))

                if confirm_action:
                    pygame.draw.circle(unit.map.engine.surface, (0, 0, 0), (450, 528), 5)
                else:
                    pygame.draw.circle(unit.map.engine.surface, (0, 0, 0), (450, 543), 5)


                # Checks if there is a unit underneath the cursor, and if there is, plots the unit's data.
                selected = False
                selected = unit.map.cursor_key_search()
                if selected is not False:
                    unit.map.all_units_by_name[selected].plot_stats()

                pygame.display.flip()
                unit.map.engine.clock.tick(60)


class AirLift(Skill):

    def __init__(self):
        """

        function name: __init__

        purpose: initializes the Skill and defines relevant variables

        """


        desc = "Tengu Air Lift"
        name = "Air Lift"
        minimum = 0
        cost = 0
        self.minrange = 1
        self.maxrange = 1
        Skill.__init__(self, name, desc, minimum, cost)


    def check_usability(self, unit):
        """
        Function name: check_usability:

        A unit can only use this skill if they have the trait property Flight and has not moved yet.
        """


        if unit.has_trait_property('Flight') and not unit.moved:
            return True
        else:
            return False


    def check_criteria(self, unit, target):
        """
        Function name: check_criteria: checks if a unit can be picked up.

        Inputs: unit, target
                unit - unit performing this action
                target - unit attempting to be picked up

        Checks whether unit is on the same team and is within the minimum specified range.


        """

        if target.team == unit.team:
            distance = (target.location_tile - unit.location_tile).get_magnitude()
            if distance >= self.minrange and distance <= self.maxrange:
                return True
            else:
                return False
        else:
            return False

    def player_interface(self, unit):
        """
        Function name: Player_interface

        Purpose: Method called in order for the player to carry out this skill.

        inputs: unit - unit performing this action

        output: T/F - whether this action was executed (T) or cancelled (F)
                EXP - how much EXP this unit gets

        """

        # A unit performing this action gains a flat 10 EXP / action
        EXP = 10

        skill_range = self.generate_skill_range(self.minrange, self.maxrange)
        drop_range = self.generate_move_range(unit, 1)

        while True:


            passenger = self.select_unit(unit, skill_range, 'healing')

            if passenger:


                unit_old = unit.location_tile.copy()
                passenger_old = passenger.location_tile.copy()

                passenger.move_to(unit.location_tile)
                # Hides passenger temporarily
                passenger.update_location(-1, -1)

                unit.get_moves_path()

                move_aborted = False

                while not move_aborted:

                    destination = self.select_move(unit, unit.validmoves)

                    if destination:
                        unit.move_to(destination)

                        drop_aborted = False
                        while not drop_aborted:
                            drop_destination = self.select_move(unit, drop_range)
                            if drop_destination:
                                passenger.update_location(*unit.location_tile)
                                passenger.move_to(drop_destination)
                                return True, EXP

                            else:
                                unit.move_to(unit_old)
                                drop_aborted = True

                    else:
                        unit.update_location(*unit_old)
                        passenger.update_location(*unit_old)
                        passenger.move_to(passenger_old)
                        move_aborted = True

            else:
                return False, 0


class DizzyAttack(Skill):

    def __init__(self):


        """
        function name: __init__


        purpose: initializes the Skill and defines relevant variables

        """


        desc = "Dizzy Attack test"
        name = "Dizzy Attack"
        minimum = 400
        cost = 100

        Skill.__init__(self, name, desc, minimum, cost)

        self.star_image = [pygame.image.load(os.path.join('images', 'bullets', 'star_blue.png')).convert_alpha(),
                      pygame.image.load(os.path.join('images', 'bullets', 'star_teal.png')).convert_alpha(),
                      pygame.image.load(os.path.join('images', 'bullets', 'star_red.png')).convert_alpha(),
                      pygame.image.load(os.path.join('images', 'bullets', 'star_green.png')).convert_alpha(),
                      ]



    def animation(self, unit, target):
        """
        function name: animation


        purpose: Draws stars exploding over the target.
        """

        star_group = pygame.sprite.RenderUpdates()

        num_stars = 5
        for i in xrange(0,num_stars):
            start_coord = target.location_pixel + Vector2(17,0) - 35* unit.map.screen_shift
            # Stars emtited from a 90 degree angle above target.
            angle = -1*(90-randint(-20,20))*3.14/180
            speed = 5
            velocity = speed*Vector2(cos(angle), sin(angle))

            star_group.add(StarSprite(choice(self.star_image), start_coord, velocity))


        # First Frame

        unit.map.render_background()
        unit.map.render_all_units()
        unit.map.engine.surface.blit(unit.map.engine.menu_board, (0, 490))
        unit.map.engine.surface.blit(unit.map.engine.map_spell_board, (175, 0))
        unit.plot_stats()

        bg_surface = unit.map.engine.surface.copy()

        star_group.clear(unit.map.engine.surface, bg_surface)
        star_group.draw(unit.map.engine.surface)

        pygame.display.flip()
        unit.map.engine.clock.tick(60)


        for tick in xrange(0,60):

            star_group.clear(unit.map.engine.surface, bg_surface)
            star_group.update()

            rects = star_group.draw(unit.map.engine.surface)

            pygame.display.update(rects)
            unit.map.engine.clock.tick(60)



    def player_interface(self, unit):


        """
        Function name: Player_interface

        Purpose: Method called in order for the player to carry out this skill.

        inputs: unit - unit performing this action

        output: T/F - whether this action was executed (T) or cancelled (F)
                EXP - how much EXP this unit gets

        """

        target_list = self.select_nearby_units(unit, 'enemy', 2)
        target_list = [target for target in target_list if not target.invincible]
        skill_range = self.generate_skill_range(0, 2)
        confirm_action = self.confirm_proximity_action(unit, target_list, skill_range, 'attack')

        # 10EXP is awarded for each unit affected.
        EXP = 0

        if confirm_action:
            # User of this skill can also become dizzy
            target_list.append(unit)

            # 75% of making target dizzy
            for target in target_list:
                roll = randint(0, 100)
                if roll > 25 and 'Dizzy' not in target.status:

                    unit.map.engine.sfx_system.sound_catalog['hit'].play()
                    self.animation(unit, target)
                    target.give_status('Dizzy')
                    EXP += 10


            return True, EXP

        else:
            return False, 0

class StarSprite(pygame.sprite.Sprite):

    def __init__(self, image, position, velocity):
        """
        Sprite for star images used in Dizzy Attack's animation
        """

        # Gravity exerts a downward acceleration.
        self.acceleration = Vector2(0,0.3)

        self.velocity = velocity
        self.image = image
        self.floor = min(position.y+50, 490)
        self.rect = self.image.get_rect()
        self.float_position = position
        self.rect.center = (int(self.float_position.x),int(self.float_position.y))

        pygame.sprite.Sprite.__init__(self)

    def update(self):

        """
        Does an Euler's method update on position based on acceleration / velocity every time function is called.
        """

        self.velocity += self.acceleration
        self.float_position += self.velocity

        self.rect.center = (int(self.float_position.x),int(self.float_position.y))

        if self.rect.bottom > self.floor:
            self.kill()

class CatsCurse(Skill):

    def __init__(self):

        desc = "Cats Curse test"
        name = "Cat\'s Curse"
        minimum = 100
        cost = 100

        Skill.__init__(self, name, desc, minimum, cost)

    def player_interface(self, unit):

        """
        Function name: Player_interface

        Purpose: Method called in order for the player to carry out this skill.

        inputs: unit - unit performing this action

        output: T/F - whether this action was executed (T) or cancelled (F)
                EXP - how much EXP this unit gets

        """

        target_list = self.select_nearby_units(unit, 'enemy', 2)

        targets_with_pos_status = []
        for target in target_list:
            if target.status and not target.invincible:
                for status_name in target.status.keys():
                    if status_name != "High Spirit" and unit.map.engine.status_effects_catalog[status_name].positive_status:
                        targets_with_pos_status.append(target)
                        break


        # 10EXP is awarded for each unit affected.
        EXP = 10*len(targets_with_pos_status)


        skill_range = self.generate_skill_range(0, 2)
        confirm_action = self.confirm_proximity_action(unit, targets_with_pos_status, skill_range, 'attack')

        if confirm_action:
            # Erase all positive status effects on enemy units
            for target in targets_with_pos_status:
                for status_name in target.status.keys():
                    if unit.map.engine.status_effects_catalog[status_name].positive_status:
                        target.remove_status(status_name)

            return True, EXP

        else:
            return False, 0


class IllusionField(Skill):

    def __init__(self):

        desc = "Illusion Field"
        name = "Illusion Field"
        minimum = 400
        cost = 100

        Skill.__init__(self, name, desc, minimum, cost)

    def player_interface(self, unit):

        target_list = self.select_nearby_units(unit, 'ally', 2)

        target_list.remove(unit)
        skill_range = self.generate_skill_range(0, 2)
        confirm_action = self.confirm_proximity_action(unit, target_list, skill_range, 'healing')


        # 10EXP is awarded for each unit affected.
        EXP = 10*len(target_list)

        if confirm_action:


            # Erase all positive status effects on enemy units
            for target in target_list:
                if 'Illusion Veil' not in target.status.keys():

                    unit.map.engine.sfx_system.sound_catalog['support1'].play()
                    unit.map.show_animation('magic_cast',target.location_tile)
                    target.give_status("Illusion Veil")

            return True, EXP

        else:
            return False, 0

class FocusedMovement(Skill):

    def __init__(self):

        desc = "focused movement test"
        name = "Focused Movement"
        minimum = 0
        cost = 0

        Skill.__init__(self, name, desc, minimum, cost)

    def check_usability(self, unit):
        """
        Focused Movement costs 0 points
        """

        return not unit.moved

    def player_interface(self, unit):

        # No EXP is awarded for this action

        unit.focused = True
        MOVEMENT_MULTIPLIER = 0.5

        # Temporarily stores current stats
        current_moves = unit.moves
        unit.moves = int(current_moves*MOVEMENT_MULTIPLIER)

        unit.get_moves_path()
        unit.move_loop()

        # Restores unit to original states
        unit.update_stats()
        unit.moves = current_moves
        unit.get_moves_path()
        unit.focused = False

        return False, 0


class Teleport(Skill):

    def __init__(self):

        desc = "teleport test"
        name = "Teleport"
        minimum = 300
        cost = 75

        Skill.__init__(self, name, desc, minimum, cost)

    def check_usability(self, unit):
        """
        Function: Check Usability
        Purpose: Checks if the unit has enough spirit charge for one use of teleport
        """

        if unit.spirit >= self.sc_cost and unit.spirit >= self.sc_minimum and not unit.moved:
            return True
        else:
            return False

    def player_interface(self, unit):
        """
        Player selects a location to teleport to. Ends turn.
        """
        # No EXP is gained from this action

        current_position_x, current_position_y = (unit.location_tile.x, unit.location_tile.y)
        move_range = self.generate_move_range(unit, unit.moves+2)
        new_pos = self.select_move(unit, move_range)
        if new_pos:

            unit.map.engine.fade_to('white', 0.1)
            unit.update_location(new_pos.x, new_pos.y)
            unit.map.engine.fade_from('white', 0.1)
            unit.moved = True
            unit.menu_loop(can_move = False, can_act = True)

            if unit.turnend:
                return True, 0
            else:
                unit.update_location(current_position_x, current_position_y)
                unit.map.cursor_pos.x, unit.map.cursor_pos.y = current_position_x, current_position_y
                unit.moved = False
                return False, 0
        else:
            return False, 0



class DoubleAction(Skill):

    def __init__(self):

        desc = "test"
        name = "Double Action"
        minimum = 500
        cost = 150
        self.minrange = 2
        self.maxrange = 6

        Skill.__init__(self, name, desc, minimum, cost)

    def check_usability(self, unit):
        """
        Checks condition for using
        """
        if unit.spirit >= self.sc_minimum:
            return True

    def check_criteria(self, unit, target):
        """
        Checks whether an opposing team's unit is in range
        """

        if target.team != unit.team:
            distance = (target.location_tile - unit.location_tile).get_magnitude()
            if distance >= self.minrange and distance <= self.maxrange:
                return True
            else:
                return False
        else:
            return False

    def player_interface(self, unit):


        # No EXP is gained from this action

        action_executed = unit.spell_loop_a()
        if action_executed:
            action_executed = False
            # Once an action has been committed to, force unit to select a second move.
            while not action_executed:
                action_executed = unit.spell_loop_a()
            return True, 0
        else:
            return False, 0

class SummonDoll(Skill):

    def __init__(self):

        desc = "Summons dolls to the field"
        name = "Summon Doll"
        minimum = 400
        cost = 100
        self.minrange = 1
        self.maxrange = 2
        Skill.__init__(self, name, desc, minimum, cost)

    def check_usability(self, unit):
        return not unit.moved and unit.proxy_units < 2 and unit.spirit >= self.sc_cost and unit.spirit >= self.sc_minimum

    def player_interface(self, unit):
        """
        Player selects a location to summon to. Ends turn.
        """


        # No EXP is gained from this action

        current_position_x, current_position_y = (unit.location_tile.x, unit.location_tile.y)

        # Can only summon on empty square. I am hijacking move methods for this
        # Additionally, this prevents Alice from summoning a doll behind an enemy
        summon_range = self.generate_skill_range(1, 1)
        summon_pos  = self.select_move(unit, summon_range)

        if summon_pos:

            if unit.proxy_units is 0:
                unit_name = "Shanghai"
            else:
                unit_name = "Hourai"

            try:
                # Add to party
                # (copies from mapaction.add_temporary_ally

                doll = unit.map.engine.player_units_catalog[unit_name]

                #Note: for somereason, I don't have any attack abilities.
                unit.map.store_unit(doll, 1)

                # set doll to same level as Alice
                doll.level = unit.level
                doll.update_stats()
                doll.HP = doll.maxHP


                # Summoned unit does not act on same turn
                doll.turnend = False

                # Set Doll unit's parent unit
                doll.set_parentunit(unit)

                # Increment proxy_units by 1
                unit.proxy_units += 1

                print "Summoned %s!" % unit_name
            except KeyError:
                print "ERROR: Unit name mismatch in add to party subaction:", unit_name

            # Update summoned doll location
            try:
                unit.map.all_units_by_name[unit_name].update_location(*summon_pos)
            except KeyError:
                print "ERROR: Unit name mismatch in set unit pos subaction", unit_name

            # Update summoned doll skills
            spell = "Fireball"
            try:
                unit.map.all_units_by_name[unit_name].add_spell(unit.map.engine.spell_catalog[spell].construct_spell())
            except KeyError:
                print "ERROR: Unit name mismatch in assign_spell subaction", unit_name, spell

            unit.map.all_units_by_name[unit_name].add_trait(unit.map.engine.trait_catalog['Self Destruct'])

            return True, 0
        else:
            return False, 0

class MagicFortress(Skill):

    def __init__(self):
        desc = 'Test'
        name = "Magic Fortress"
        minimum = 300
        cost = 50
        Skill.__init__(self, name, desc, minimum, cost)


    def player_interface(self, unit):

        # 10 EXP is gained for this action
        EXP = 10

        target_list = [unit]
        skill_range = []
        confirm_action = self.confirm_proximity_action(unit, target_list, skill_range, 'healing')
        if confirm_action:

            unit.map.engine.sfx_system.sound_catalog['support1'].play()
            unit.map.show_animation('barrier_spell', unit.location_tile)

            unit.give_status('Magic Fortress')
            return True, 10
        else:
            return False, 10

class FieldMedic(Skill):

    def __init__(self):

        desc = 'test'
        name = "Field Medic"
        minimum = 400
        cost = 100
        Skill.__init__(self, name, desc, minimum, cost)

    def player_interface(self, unit):


        target_list = self.select_nearby_units(unit, 'ally', 2)
        target_list = [target for target in target_list if target.HP < target.maxHP]

        skill_range = self.generate_skill_range(0, 2)
        confirm_action = self.confirm_proximity_action(unit, target_list, skill_range, 'healing')


        # 10 EXP is gained per unit affected
        EXP = 10*len(target_list)

        if confirm_action:
            # User of this skill can also become dizzy
            target_list.append(unit)
            for target in target_list:
                if target.HP < target.maxHP:

                    # Heals 15% of target's max HP

                    unit.map.center_on(target)
                    starting_HP = target.HP
                    healing_amount = int(0.15*target.maxHP)

                    unit.map_heal(target, 'Field Medic', healing_amount)

            return True, EXP

        else:
            return False, EXP

class DispelHistory(Skill):

    def __init__(self):

        desc = 'test'
        name = "Dispel History"
        minimum = 400
        cost = 100
        Skill.__init__(self, name, desc, minimum, cost)

    def player_interface(self, unit):

        target_list = self.select_nearby_units(unit, 'ally', 1)

        targets_with_neg_status = []
        for target in target_list:
            if target.status:
                for status_name in target.status.keys():
                    if unit.map.engine.status_effects_catalog[status_name].type != "Other" and not unit.map.engine.status_effects_catalog[status_name].positive_status:
                        targets_with_neg_status.append(target)
                        break

        skill_range = self.generate_skill_range(0, 1)
        confirm_action = self.confirm_proximity_action(unit, targets_with_neg_status, skill_range, 'healing')


        # 10 EXP is gained per unit affected
        EXP = 10*len(targets_with_neg_status)

        if confirm_action:

            unit.map.engine.sfx_system.sound_catalog['heal'].play()
            unit.map.show_animation('healing_spell',unit.location_tile)

            # Removes all negative status effects of nearby allies
            target_list.append(unit)
            for target in target_list:
                for status in target.status.keys():
                    if not unit.map.engine.status_effects_catalog[status].positive_status and status != 'Low Spirit':
                        target.remove_status(status)

            return True, EXP

        else:
            return False, EXP

class UnleashGrimoire(Skill):

    def __init__(self):

        desc = "test"
        name = "Unleash Grimoire"
        minimum = 500
        cost = 500
        self.minrange = 2
        self.maxrange = 6

        Skill.__init__(self, name, desc, minimum, cost)

    def check_usability(self, unit):
        """
        Checks condition for using
        """
        if unit.spirit >= self.sc_cost and unit.spirit >= self.sc_minimum and not unit.moved:
            return True
        else:
            return False


    def player_interface(self, unit):

        # No EXP gained from using this action

        unit.give_status('Mega Offense')

        action_executed = unit.spell_loop_a()
        if action_executed:

            unit.remove_status('Mega Offense')
            return True, 0
        else:

            unit.remove_status('Mega Offense')
            return False, 0

class Purify(Skill):


    def __init__(self):

        """
        Removes all negative status effects from user
        """

        desc = "test"
        name = "Purify"
        minimum = 400
        cost = 50
        self.minrange = 2
        self.maxrange = 6

        self.yyorb_image = pygame.image.load(os.path.join('images','bullets', 'yyorb_red.png')).convert_alpha()
        
        Skill.__init__(self, name, desc, minimum, cost)


    def check_usability(self, unit):
        """
        Checks condition for using
        """

        # Checks that unit has enough SC and has an existing negative status effect
        if unit.spirit >= self.sc_cost and unit.spirit >= self.sc_minimum:
            for status_name in unit.status.keys():
                if not unit.map.engine.status_effects_catalog[status_name].positive_status and  unit.map.engine.status_effects_catalog[status_name].type != "Other":
                    return True
            else:
                return False
        else:
            return False

    def animation(self, unit):

        """
        name: animation
        Purpose: draws two Ying Yang orbs orbiting around Reimu
        """

        # Origin
        x_o = (unit.location_pixel.x+35/2-self.yyorb_image.get_width()/2)
        y_o = (unit.location_pixel.y+35/2-self.yyorb_image.get_height()/2)

        for i in xrange(0, 60):

            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()

            unit.map.render_background()
            unit.map.render_all_units()
            unit.map.render_cursor()
            unit.map.engine.surface.blit(unit.map.engine.menu_board, (0, 490))
            unit.map.engine.surface.blit(unit.map.engine.map_spell_board, (175, 0))
            unit.plot_stats()

            # parametric equations for circular orbit, orbitting twice over the 60 frame animation
            x1 = x_o+(60-i)*sin(i*3.14/15)
            y1 = y_o+(60-i)*cos(i*3.14/15)
            x2 = x_o+(60-i)*sin(3.14+i*3.14/15)
            y2 = y_o+(60-i)*cos(3.14+i*3.14/15)


            position_vector1 = Vector2(x1,y1)-35*unit.map.screen_shift
            position_vector2 = Vector2(x2,y2)-35*unit.map.screen_shift

            unit.map.engine.surface.blit(self.yyorb_image, (int(position_vector1.x), int(position_vector1.y)))
            unit.map.engine.surface.blit(self.yyorb_image, (int(position_vector2.x), int(position_vector2.y)))
            pygame.display.flip()
            unit.map.engine.clock.tick(60)


    def player_interface(self, unit):

        range = self.generate_skill_range(0, 0)

        # 10 EXP is awarded
        EXP = 10

        # Cures all negative status effects other than Low Spirit
        if self.confirm_proximity_action(unit, [unit], range, 'healing'):
            unit.map.engine.sfx_system.sound_catalog['support3'].play()
            self.animation(unit)
            for status_name in unit.status.keys():
                if not unit.map.engine.status_effects_catalog[status_name].positive_status and status_name != "Low Spirit":
                    unit.remove_status(status_name)

            return True, EXP

        else:
            return False, 0

class SelfDestruct(Skill):

    def __init__(self):

        """
        Blows up and does damage to adjacent units.
        """

        desc = "test"
        name = "Self Destruct"
        minimum = 0
        cost = 0
        self.maxrange = 1

        Skill.__init__(self, name, desc, minimum, cost)

    def calculate_damage(self, unit, target):

        """
        Calculates the damage done to target
        """

        # Base effect is based on unit's magic stat
        base_effect = 4
        effect = unit.MAG*base_effect

        # Shield uses standard defense calculation for spell actions against physical defense
        target_total_mods = target.compute_total_stat_mods()
        shield = (target.DEF*(1.0+target_total_mods[1]) + target.spell_actions[target.equipped].defmod)*target.map.terrainmap[tuple(target.location_tile)][0].TDEF*target.spell_actions[target.equipped].shield

        # Final damage = Effect - Shield
        damage = int(effect - shield)

        return damage

    def player_interface(self, unit):


        target_list = self.select_nearby_units(unit, 'both', 1)

        # Does not work on units with invincibility barrier.
        target_list = [target for target in target_list if not target.invincible]

        target_list.remove(unit)

        skill_range = self.generate_skill_range(1, 1)
        confirm_action = self.confirm_proximity_action(unit, target_list, skill_range, 'attack')



        if confirm_action:

            unit.map.engine.sfx_system.sound_catalog['explode'].play()
            unit.map.engine.fade_to('white',1.0)
            unit.map.kill(unit)

            unit.map.render_background()
            unit.map.render_all_units()
            unit.map.render_cursor()
            unit.map.render_menu_panel()

            unit.map.engine.fade_from('white',1.0)

            # Erase all positive status effects on enemy units
            for target in target_list:
                damage = self.calculate_damage(unit, target)
                old_hp = target.HP
                target.HP -= damage
                if target.HP < 0:
                    target.HP = 0


                # Draw the effect of each target getting hit

                unit.map.engine.sfx_system.sound_catalog['hit'].play()
                effect_text = unit.map.engine.render_outlined_text(str(damage), unit.map.engine.cfont, (255, 0, 0), (255, 255, 255))

                unit.map.render_background()
                unit.map.render_all_units()
                unit.map.render_cursor()
                unit.map.engine.surface.blit(unit.map.engine.menu_board, (0, 490))
                target.plot_stats()
                unit.map.engine.surface.blit(effect_text, ((target.location_pixel.x+18-effect_text.get_width()/2, target.location_pixel.y-25)-unit.map.screen_shift*unit.map.engine.tilesize))

                pygame.display.flip()
                unit.map.engine.clock.tick(60)
                unit.map.engine.pause(1)


                target.render_hp_change(old_hp, target.HP)

                if target.HP == 0:
                    target.alive = False
                    unit.map.kill(target, True)

            if unit.is_proxy_unit and unit.parentunit.alive:
                EXP = 0
                for target in target_list:
                    if target.team == 2:
                        # 10 EXP for target still alive
                        if target.alive:
                            EXP += 10
                        # 30 EXP for target defeated
                        else:
                            EXP += 30

                unit.parentunit.exp += EXP
                level_up = unit.parentunit.level_up_check()

                unit.parentunit.plot_results(unit.parentunit, EXP, level_up, 0, False, 0, 0, 0, 0)

            return True, 0

        else:
            return False, 0